#!/usr/bin/perl -w

package DistroMap;
use strict;
use warnings;

our $VERSION = 0.12;

my %default_branch = qw/
altlinux	sisyphus
fedora		rawhide
generic		default
jpackage	default
/;

# TODO: change defaults to generic:generic
# and assure that quick initialization is done if generic
our $default_originrepo='generic';
our $default_originbranch=undef;
our $default_destrepo='generic';
our $default_destbranch=undef;

our ($opt_origin_check_distrodb, $opt_dest_check_distrodb);

our $flag_longopt_published;
our @LONGOPT=(
    'originrepo=s' => \$default_originrepo,
    'originbranch=s' => \$default_originbranch,
    'origin-check-distrodb=f' => \$opt_origin_check_distrodb,

    'destrepo=s' => \$default_destrepo,
    'destbranch=s' => \$default_destbranch,
    'dest-check-distrodb=f' => \$opt_dest_check_distrodb,
    );

my $SMAP;
my $SMAPBACK;
my $BMAP;
my $EXACT_GMAP;
my $APPROX_GMAP;
my $VERSION_UNCOMPARABLE;
my $verbose=0;
my $eterbuild='/usr/share/eterbuild/pkgrepl';

sub local_2alt_dir {
    my $repo=shift;
    return $ENV{HOME}.'/src/repo/distromap-'.$repo.'2alt.git';
}

# public static
sub longopt_usage {
    print "DistroMap specific options:\n";
    print "	--originrepo <origin repository name>.".($default_originrepo ? " Default is ".$default_originrepo : ''),"\n";
    print "	--originbranch <origin repository branch name>.",($default_originbranch ? " Default is ".$default_originbranch : ''),"\n";
    print "	--destrepo <destination repository name>.",($default_destrepo ? " Default is ".$default_destrepo : ''),"\n";
    print "	--destbranch <destination repository branch name>.",($default_destbranch ? " Default is ".$default_destbranch : ''),"\n";
    print "	--{origin,dest}-check-distrodb <days>. 
Warn if DistroDB is older than <days> days. Default is ".
($opt_origin_check_distrodb ? "$opt_origin_check_distrodb days" : 'do not check')." for origin and ";
($opt_dest_check_distrodb ? "$opt_dest_check_distrodb days" : 'do not check')." for dest.\n";

}

sub new {
    my $proto=shift;
    my $class=ref($proto) || $proto;
    my $self = {
	'-originrepo' => $default_originrepo,
	'-originbranch' => $default_originbranch,
	'-destrepo' => $default_destrepo,
	'-destbranch' => $default_destbranch,
	-origin_check_distrodb => $opt_origin_check_distrodb,
	-dest_check_distrodb => $opt_dest_check_distrodb,
	@_,
    };
    bless($self,$class);

    # force lowercase of repo and branch names
    foreach (
	'-originrepo',
	'-originbranch',
	'-destrepo',
	'-destbranch',
	) {
	my $val=$self->{$_};
	$self->{$_}=lc($val) if $val;
    }

    $self->{'-originbranch'}=&DistroMap::Repository::select_branch($self->{'-originrepo'},$self->{'-originbranch'});
    $self->{'-destbranch'}=&DistroMap::Repository::select_branch($self->{'-destrepo'},$self->{'-destbranch'});
    $self->{ORIGIN}=DistroMap::Repository->new(
	-repo   => $self->{-originrepo},
	-branch => $self->{-originbranch},
	-check_distrodb => $self->{-origin_check_distrodb},
	);
    $self->{DEST}=DistroMap::Repository->new(
	-repo   => $self->{-destrepo},
	-branch => $self->{-destbranch},
	-check_distrodb => $self->{-dest_check_distrodb},
	);
    unless ($SMAP) {
	$self->__load();
    }
    return $self;
}

sub origin {
    my ($self)=@_;
    return $self->{ORIGIN};
}

sub dest {
    my ($self)=@_;
    return $self->{DEST};
}

sub unimap_source_dest2origin {
    my ($self,$name)=@_;
    my $map=$SMAPBACK->{$name};
    return unless $map;
    return $map->[0];
}

sub unimap_source_origin2dest {
    my ($self,$name)=@_;
    my $map=$SMAP->{$name};
    return unless $map;
    return $map->[0];
}

sub multimap_binary_origin2dest {
    my ($self,$name)=@_;
    my $map=$BMAP->{$name};
    return @$map if $map;
    my $dest=$self->dest;
    my $smap=$self->__stem_map($name);
    return $smap if $smap and $dest->exist_provides($smap);
    if ($name=~s/-devel$//) {
	$smap=$self->__stem_map($name);
	return $smap.'-devel' if $smap and $dest->exist_provides($smap.'-devel');
	return 'lib'.$name.'-devel' if $dest->exist_provides('lib'.$name.'-devel');
	if ($name=~s/-libs$// && $dest->exist_provides('lib'.$name.'-devel')) {
	    return 'lib'.$name.'-devel';
	}
    } elsif ($name=~s/-docs?$//) {
	$smap=$self->__stem_map($name);
	for my $prefix ('','lib') {
	    for my $suffix (qw/-doc -docs/) {
		return $prefix.$smap.$suffix if $smap and $dest->exist_provides($prefix.$smap.$suffix);
	    }
	}
    }
    return;
}

sub unimap_exact_group_origin2dest {
    my ($self,$name)=@_;
    return $EXACT_GMAP->{$name};
}

sub unimap_approximate_group_origin2dest {
    my ($self,$name)=@_;
    return $APPROX_GMAP->{$name};
}

sub versions_are_uncomparable {
    my ($self,$name)=@_;
    return $VERSION_UNCOMPARABLE->{$name};
}


package DistroMap::Repository;

use strict;
use warnings;

our $opt_check_distrodb;

our @LONGOPT=(
    'check-distrodb=f' => \$opt_check_distrodb,
    );

my $REPODATA={};

# public static
sub longopt_usage {
    print "DistroDB specific options:\n";
    print "	--check-distrodb <days>. Warn if DistroDB is older than <days> days. Default is ".($opt_check_distrodb ? "$opt_check_distrodb days" : 'do not check').".\n";
}

sub new {
    my $proto=shift;
    my $class=ref($proto) || $proto;
    my $self = {
	-repo => 'generic',
	-branch => undef,
	-check_distrodb => $opt_check_distrodb,
	@_,
    };
    bless($self,$class);

    # force lowercase of repo and branch names
    foreach (
	'-repo',
	'-branch',
	) {
	my $val=$self->{$_};
	$self->{$_}=lc($val) if $val;
    }

    $self->{'-branch'}=&select_branch($self->{'-repo'},$self->{'-branch'});
    print STDERR "DistroMap::Repository: new ",$self->{'-repo'},", ",$self->{'-branch'},"\n" if $verbose and $verbose > 2;

    # Live Fast Die Young - warn if repo is older than -check_distrodb days
    if ($self->{-check_distrodb}) {
	my $repo=$self->{'-repo'};
	my $branch=$self->{'-branch'};
	my $cachedir=$ENV{'HOME'}."/.cache/distrodb/$repo/$branch";
	my $systemdir="/usr/share/distrodb/$repo/$branch";
	my $dir=&DistroMap::__select_most_fresh_dir($cachedir, $systemdir);
	if (not defined $dir or not -e $dir) {
	    die "FATAL: DistroMap repository data for $repo:$branch is missing!
run distromap-update-repocop-db-$repo-$branch!\n";
	} elsif (-M $dir > $self->{-check_distrodb}) {
	    warn "WARNING: repository data for $repo:$branch is too old. run distromap-update-repocop-db-$repo-$branch\n";
	}
    }
    $self->{'DATA'}=&__get_repodata($self->{'-repo'},$self->{'-branch'});
    return $self;
}

sub select_branch {
    my ($repo, $branch)=@_;
    return $branch if defined $branch;
    $branch = $default_branch{$repo};
    return $branch if defined $branch;
    return 'default';
}

sub translate {
    my ($self,$db,$name)=@_;
    my $map=$self->_get_db($db)->{$name};
    return unless $map;
    return wantarray ? @$map : $map->[0];
}

sub libname2binpkgname {
    my ($self,$name)=@_;
    my $map=$self->_get_db('devel-libs')->{$name};
    return unless $map;
    return wantarray ? @$map : $map->[0];
}

sub header2binpkgname {
    my ($self,$name)=@_;
    my $map=$self->_get_db('headers')->{$name};
    return unless $map;
    return wantarray ? @$map : $map->[0];
}

sub pkgconfigname2binpkgname {
    my ($self,$name)=@_;
    my $map=$self->_get_db('pkg-config')->{$name};
    return unless $map;
    return wantarray ? @$map : $map->[0];
}

sub path2binpkgname {
    my ($self,$name)=@_;
    my $map=$self->_get_db('path')->{$name};
    return unless $map;
    return wantarray ? @$map : $map->[0];
}

sub exist_provides {
    my ($self,$name)=@_;
    return $self->_get_db('provides')->{$name};
}

# my @defpath=qw!/bin/ /usr/bin/ /sbin/ /usr/sbin/!;
# sub find_in_path {
#     my ($self,$name,$pathref)=@_;
#     my $pathdb=$self->_get_db('path');
#     foreach my $dir ($pathref ? @$pathref:@defpath) {
# 	my $file= $pathdb->{$dir.$name};
# 	return $file if defined $file;
#     }
#     return;
# }

#package DistroMap::Repository::LazyLoader;


########## PRIVATE #################
package DistroMap::Repository;

sub __get_repodata {
    my @path=@_;
    my $node=$REPODATA;
    while (@path>0) {
	my $step=shift @path;
	$node->{$step}={} unless defined $node->{$step};
	$node=$node->{$step};
    }
    return $node;
}

sub _get_db {
    my ($self,$type)=@_;
    my $MAP=$self->{'DATA'}->{$type};
    return $MAP if $MAP;
    $MAP={};
    $self->{'DATA'}->{$type}=$MAP;
    $self->__load_type($type,$MAP);
    return $MAP;
}

sub __load_type {
    my ($self,$type,$MAP)=@_;
    my $repo=$self->{'-repo'};
    return if $repo eq 'generic';
    my $branch=$self->{'-branch'};
    print STDERR "DistroMap: reading $repo $branch $type\n" if $verbose and $verbose > 1;
    foreach my $file (&__glob_1map_files($repo,$branch,$type)) {
	&DistroMap::__load_map_as($file,$MAP);
    }
}

sub __glob_1map_files {
    my ($repo,$branch,$type)=@_;
    my @out;
    my $cachedir=$ENV{'HOME'}."/.cache/distrodb/$repo/$branch/$type";
    my $systemdir="/usr/share/distrodb/$repo/$branch/$type";
    my $dir=&DistroMap::__select_most_fresh_dir($cachedir, $systemdir);
    push @out, glob "$dir/*.{txt,tsv}" if $dir;
    return @out;
}

########## END PRIVATE #################


########## PRIVATE #################
package DistroMap;

sub __stem_map {
    my ($self,$name)=@_;
    my $map=$SMAP->{$name};
    return $map->[0] if $map and $map->[0] ne $name;
    return;
}

sub __load {
    my ($self)=@_;
    $SMAP={};
    $BMAP={};
    $BMAP={};
    $EXACT_GMAP={};
    $APPROX_GMAP={};
    $VERSION_UNCOMPARABLE={};
    my $originrepo=$self->{'-originrepo'};
    my $destrepo=$self->{'-destrepo'};
    my $originbranch=$self->{'-originbranch'};
    my $destbranch=$self->{'-destbranch'};

    return if $originrepo eq 'generic' and $destrepo eq 'generic';
    return if $originrepo eq $destrepo and $originbranch eq $destbranch;

    foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,'source')) {
	&__load_map_as($file,$SMAP,$SMAPBACK);
    }
    foreach my $file ($self->__find_2map_files($destrepo,$destbranch,$originrepo,$originbranch,'source')) {
	&__load_map_as($file,$SMAPBACK);
    }
    foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,'binary')) {
	&__load_map_as($file,$BMAP);
    }

    foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,'group-strict')) {
	&__load_unimap_as($file,$EXACT_GMAP);
    }
    foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,'group-approx')) {
	&__load_unimap_as($file,$APPROX_GMAP);
    }

    foreach my $file ($self->__find_2map_files($originrepo,$originbranch,$destrepo,$destbranch,'noversion')) {
	&__load_list_as($file,$VERSION_UNCOMPARABLE);
    }
}

sub __eabbr {
    my $a=shift;
    return $a;
}

sub __etype {
    my $a=shift;
    return 'pkgrepl' if $a eq 'binary';
    return 'grprepl' if $a eq 'group-approx';
    return;
}

sub __find_2map_files {
    my ($self, $originrepo,$originbranch,$destrepo,$destbranch,$type)=@_;
    my @out;
    # TODO: implement chains
    if ($originrepo eq 'jpackage') {
	push @out, $self->__glob_2map_files('generic','default',$destrepo,$destbranch,$type);
	push @out, $self->__glob_2map_files('fedora','rawhide',$destrepo,$destbranch,$type);
    } elsif ($originrepo eq 'fedora' or $originrepo eq 'suse') {
	push @out, $self->__glob_2map_files('generic','default',$destrepo,$destbranch,$type);
    }

    my $origindef=$DistroMap::default_branch{$originrepo};
    my $destdef=$DistroMap::default_branch{$destrepo};
    push @out, $self->__glob_2map_files($originrepo,$origindef,$destrepo,$destdef,$type)
	if $origindef and $origindef ne $originbranch and $destdef and $destdef ne $destbranch;
    push @out, $self->__glob_2map_files($originrepo,$origindef,$destrepo,$destbranch,$type)
	if $origindef and $origindef ne $originbranch and (not $destdef or $destdef ne $destbranch);
    push @out, $self->__glob_2map_files($originrepo,$originbranch,$destrepo,$destdef,$type)
	if $destdef and $destdef ne $destbranch and (not $origindef or $origindef ne $originbranch);
    push @out, $self->__glob_2map_files($originrepo,$originbranch,$destrepo,$destbranch,$type);
    return @out;
}

sub __glob_2map_files {
    my ($self, $originrepo,$originbranch,$destrepo,$destbranch,$type)=@_;
    my @out;
    if ($originrepo eq 'altlinux' && $originbranch eq 'sisyphus'&& -d $eterbuild) {
	my $drepo=&__eabbr($destrepo);
	my $dbranch=&__eabbr($destbranch);
	my $etype=&__etype($type);
	push @out, grep {-f $_} "$eterbuild/$etype.$drepo", "$eterbuild/$etype.$drepo.$dbranch" if $etype;
    }
    my $local_2alt_dir=&local_2alt_dir($originrepo);
    if ($local_2alt_dir && -d $local_2alt_dir && $destrepo eq 'altlinux') {
	push @out, glob "$local_2alt_dir/$type/*.{txt,tsv}";
    } else {
	my $cachedir=$ENV{'HOME'}."/.cache/distromap/$originrepo/$originbranch/$destrepo/$destbranch/$type";
	my $systemdir="/usr/share/distromap/$originrepo/$originbranch/$destrepo/$destbranch/$type";
	my $dir=&__select_most_fresh_dir($cachedir, $systemdir);
	push @out, glob "$dir/*.{txt,tsv}" if $dir;
    }
    return @out;
}

sub __select_most_fresh_dir {
    #print STDERR "DistroMap: most_fresh_dir: args: @_\n" if $verbose and $verbose > 4;
    my @indir=grep {-d $_} @_;
    #print STDERR "DistroMap: most_fresh_dir: exists: @indir\n" if $verbose and $verbose > 4;
    return unless @indir;
    my $dir=shift @indir;
    while (scalar @indir) {
	my $newdir=shift @indir;
	$dir=$newdir if -M $dir > -M $newdir;
    }
    #print STDERR "DistroMap: most_fresh_dir: selected: $dir\n" if $verbose and $verbose > 4;
    return $dir;
}

sub __load_map_as {
    my ($file,$MAP,$MAPBACK)=@_;
    print STDERR "DistroMap: loaded map $file\n" if $verbose and $verbose > 2;
    my $split=qr'\s+';
    $split=qr' *\t+ *' if $file=~/tsv$/;
    $split=qr'|' if $file=~m!/eterbuild/!;
    open (my $fn, '<', $file) || die "can't open $file: $!";
    local $_;
    while (<$fn>) {
	chomp;
	next if /^#/;
	next if /^\s*$/;
	warn "$file: endspace found" if s/\s+$//;
	my @map=grep {$_} split (/$split/, $_);
	my $from=shift @map;
	die "$file: invalid map $_" unless @map;
	$MAP->{$from}=\@map;
	$MAPBACK->{$map[0]}=[$from] if $MAPBACK and (scalar @map == 1);
    }
    close ($fn);
}

sub __load_unimap_as {
    my ($file,$MAP,$MAPBACK)=@_;
    print STDERR "DistroMap: loaded map $file\n" if $verbose and $verbose > 2;
    my $split=qr'\s+';
    $split=qr' *\t+ *' if $file=~/tsv$/;
    $split=qr'|' if $file=~m!/eterbuild/!;
    open (my $fn, '<', $file) || die "can't open $file: $!";
    local $_;
    while (<$fn>) {
	chomp;
	next if /^#/;
	next if /^\s*$/;
	my @map=grep {$_} split (/$split/, $_);
	warn "$file: extra args in unimap: @map\n" if @map >2;
	my $from=shift @map;
	die "$file: invalid map $_" unless @map;
	$MAP->{$from}=$map[0];
	$MAPBACK->{$map[0]}=[$from] if $MAPBACK and (scalar @map == 1);
    }
    close ($fn);
}

sub __load_list_as {
    my ($file,$LIST)=@_;
    print STDERR "DistroMap: loaded list $file\n" if $verbose and $verbose > 2;
    open (my $fn, '<', $file) || die "can't open $file: $!";
    local $_;
    while (<$fn>) {
	chomp;
	next if /^#/;
	next if /^\s*$/;
	s/\s*$//;
	$LIST->{$_}=1;
    }
    close ($fn);
}

########## END PRIVATE #################


1;

=head1	NAME

DistroMap - DistroMap database query library.

=head1	DESCRIPTION

DistroMap query library 
provides an object interface to DistroMap database.

=head1	AUTHOR

Written by Igor Vlasenko <viy@altlinux.org>.

=head1	COPYING

Copyright (c) 2010 Igor Vlasenko, ALT Linux Team.

This is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

=cut

